import re

import attr
import pytest


def exact_match(full_string):
    return f"^{re.escape(full_string)}$"


@attr.dataclass(frozen=True)
class null_action:
    result: object


def test_registry(registry):
    assert registry


def test_deleted(target):
    assert target.deleted("path/to/file")


def test_not_implemented(registry):
    testing_registry = registry.Registry()
    with pytest.raises(NotImplementedError):
        testing_registry.require(None)


def test_add_target_with_non_added_parent(registry):
    testing_registry = registry.Registry()
    with pytest.raises(
        ValueError,
        match=exact_match(
            "Cannot add target 'test_target' with parent action 'action'."
            " Parent action 'action' has not been added to the registry."
        ),
    ):
        testing_registry.require(
            registry.ActionOutput("action", "test_target")
        )


def test_add_deleted_with_non_added_parent(registry, target):
    testing_registry = registry.Registry()
    target = target.deleted("path/to/file")
    with pytest.raises(
        ValueError,
        match=exact_match(
            "Cannot add target Deleted(path='path/to/file') with parent "
            "action 'action'."
            " Parent action 'action' has not been added to the registry."
        ),
    ):
        testing_registry.require(registry.ActionOutput("action", target))


def test_add_target_with_different_parent(registry):
    action = null_action(None)
    testing_registry = (
        registry.Registry()
        .require(registry.Action(action))
        .require(registry.ActionOutput(action, "test_target"))
    )
    with pytest.raises(
        ValueError,
        match=exact_match(
            "Cannot add target 'test_target' with parent action 'action'."
            " Target 'test_target' already has parent action"
            " null_action(result=None)."
        ),
    ):
        testing_registry.require(
            registry.ActionOutput("action", "test_target")
        )


def test_name_target(registry):
    testing_registry = registry.Registry()
    with pytest.raises(
        ValueError,
        match=exact_match(
            "Cannot name target 'test_target' 'test_target'."
            " Target 'test_target' has not been added to the registry."
        ),
    ):
        testing_registry.require(
            registry.TargetName("test_target", "test_target")
        )


# Does this make sense?
def test_skip_non_existent_target(registry):
    testing_registry = registry.Registry()
    with pytest.raises(
        ValueError,
        match=exact_match(
            "Cannot skip name 'test_target'."
            " Name 'test_target' has not been assigned to any target."
        ),
    ):
        testing_registry.require(registry.SkippedName("test_target"))


def test_add_deleted_via_api(api):
    for requirement in api.target(api.deleted("path"), "parent"):
        assert requirement


def test_add_action_twice(registry):
    action = null_action(None)
    testing_registry = registry.Registry().require(registry.Action(action))
    assert testing_registry == testing_registry.require(
        registry.Action(action)
    )


def test_add_input_out_of_order(registry):
    testing_registry = registry.Registry()
    with pytest.raises(
        ValueError,
        match=exact_match(
            "Cannot add target 'test_input' as an input to action"
            " 'test_action'. Action 'test_action' has not been added to the"
            " registry."
        ),
    ):
        testing_registry.require(
            registry.ActionInput("test_action", "test_input")
        )


def test_add_to_wrong_child(registry):
    action_1 = null_action(1)
    action_2 = null_action(2)
    testing_registry = (
        registry.Registry()
        .require(registry.Action(action_1))
        .require(registry.Action(action_2))
        .require(registry.ActionOutput(action_2, "test_input"))
    )
    with pytest.raises(
        ValueError,
        match=exact_match(
            "Cannot add target 'test_input' as an input to action"
            " null_action(result=1). The most recently added action is"
            " null_action(result=2)."
        ),
    ):
        testing_registry.require(registry.ActionInput(action_1, "test_input"))


def test_add_input_not_added(registry):
    action = null_action(None)
    testing_registry = registry.Registry().require(registry.Action(action))
    with pytest.raises(
        ValueError,
        match=exact_match(
            "Cannot add target 'test_target' as an input to action"
            " null_action(result=None)."
            " Target 'test_target' has not been added to the registry."
        ),
    ):
        testing_registry.require(registry.ActionInput(action, "test_target"))


def test_add_cycle(registry):
    action = null_action(None)
    testing_registry = (
        registry.Registry()
        .require(registry.Action(action))
        .require(registry.ActionOutput(action, "target"))
    )
    with pytest.raises(
        ValueError,
        match=exact_match(
            "Cannot add target 'target' as an input to action"
            " null_action(result=None)."
            " The parent action of target 'target' is"
            " null_action(result=None)."
        ),
    ):
        testing_registry.require(registry.ActionInput(action, "target"))


def test_add_twice(registry):
    action1 = object()
    action2 = object()
    action3 = object()
    requirements = [
        registry.Action(action1),
        registry.ActionOutput(action1, "target"),
        registry.Action(action2),
        registry.ActionInput(action2, "target"),
        registry.Action(action3),
    ]
    testing_registry = registry.Registry()
    for requirement in requirements:
        testing_registry = testing_registry.require(requirement)
    for requirement in requirements:
        testing_registry = testing_registry.require(requirement)



import motr._api.requirements.requirements
import motr.core.registry
import motr.core.target_name


def skipped_name(
    name: str,
) -> motr._api.requirements.requirements.Requirements:
    target_name = motr.core.target_name.coerce(name)
    yield motr.core.registry.SkippedName(target_name)



import os
import typing

import attr

CoerceToString = typing.Union[os.PathLike[str], str]


@attr.dataclass(frozen=True)
class Token:
    pass


@attr.dataclass(frozen=True)
class Deleted(Token):
    path: str


Target = typing.Union[Token, str]
CoerceToTarget = typing.Union[Token, CoerceToString]


def deleted(path: CoerceToString) -> Target:
    return Deleted(os.fspath(path))


def coerce(target: CoerceToTarget) -> Target:
    if isinstance(target, Token):
        return target
    return os.fspath(target)



import motr._api.requirements.requirements
import motr.core.registry


def build(
    requirements: motr._api.requirements.requirements.Requirements,
) -> motr.core.registry.Registry:
    registry = motr.core.registry.Registry()
    for requirement in requirements:
        registry = registry.require(requirement)
    return registry



import collections
import typing

import attr
import trio

import motr.core.exc
import motr.core.result
import motr.core.target

if typing.TYPE_CHECKING:
    import motr.core.registry

RuntimeAction = typing.Callable[
    [],
    typing.Awaitable[
        typing.Tuple[motr.core.result.Result, typing.Mapping[str, str]]
    ],
]


@attr.dataclass(frozen=True)
class TaskWrapper:
    started: trio.Event = attr.Factory(trio.Event)
    finished: trio.Event = attr.Factory(trio.Event)

    async def __call__(
        self,
        task: typing.Callable[[], typing.Awaitable[None]],
    ) -> None:
        if self.started.is_set():
            await self.finished.wait()
            return
        self.started.set()
        try:
            await task()
        finally:
            self.finished.set()


@attr.dataclass(frozen=True)
class Target:
    registry: motr.core.registry.Registry
    reporter: typing.Callable[[RuntimeAction], typing.ContextManager[None]]
    results: typing.Dict[
        motr.core.result.Result,
        typing.List[typing.Tuple[RuntimeAction, typing.Mapping[str, str]]],
    ]
    tasks: typing.DefaultDict[RuntimeAction, TaskWrapper] = attr.Factory(
        lambda: collections.defaultdict(TaskWrapper)
    )

    async def __call__(self, target: motr.core.target.Target) -> None:
        action = self.registry.parent(target)
        await self.tasks[action](Action(self, action))


@attr.dataclass(frozen=True)
class Action:
    target: Target
    action: RuntimeAction

    async def __call__(self) -> None:
        with self.target.reporter(self.action):
            async with trio.open_nursery() as nursery:
                for target in self.target.registry.inputs(self.action):
                    nursery.start_soon(self.target, target)
            result, data = await self.action()
            self.target.results[result].append((self.action, data))
            if result.aborted:
                raise motr.core.exc.MOTRTaskError(
                    f"Action {self.action!r} failed"
                )


async def run_all(
    request: typing.Callable[
        [motr.core.target.Target], typing.Awaitable[None]
    ],
    targets: typing.Iterable[motr.core.target.Target],
) -> None:
    async with trio.open_nursery() as nursery:
        for target in targets:
            nursery.start_soon(request, target)



"""
PyTest Fixtures.
"""

import enum
import pathlib
import tempfile

import pytest


@pytest.fixture(autouse=True)
def _restore_cwd(monkeypatch, base):
    monkeypatch.setattr(base, "chdir", monkeypatch.chdir)


class ToParent(enum.Enum):
    SELF = False
    PARENT = True


@pytest.fixture
def to_tmpdir(monkeypatch):
    with tempfile.TemporaryDirectory() as tmpdir:
        monkeypatch.chdir(tmpdir)
        yield tmpdir


@pytest.fixture(params=list(ToParent))
def to_parent(request):
    return request.param.value


@pytest.fixture
def implicit_expected(to_tmpdir, to_parent, monkeypatch):
    target = pathlib.Path(".")
    if to_parent:
        target /= "child"
        target.mkdir()
        monkeypatch.chdir(target)
    return pathlib.Path("motrfile.py").resolve()


@pytest.fixture
def implicit_actual(implicit_expected, to_tmpdir):
    return to_tmpdir / pathlib.Path("motrfile.py")


@pytest.fixture(scope="session", name="test_helpers")
def _test_helpers():
    import motr.test_helpers

    return motr.test_helpers


@pytest.fixture(scope="session")
def main():
    import motr.main

    return motr.main


@pytest.fixture(scope="session")
def exc():
    import motr.core.exc

    return motr.core.exc


@pytest.fixture(scope="session")
def registry():
    import motr.core.registry

    return motr.core.registry


@pytest.fixture(scope="session")
def result():
    import motr.core.result

    return motr.core.result


@pytest.fixture(scope="session")
def runner():
    import motr.core.runner

    return motr.core.runner


@pytest.fixture(scope="session")
def target():
    import motr.core.target

    return motr.core.target


@pytest.fixture(scope="session")
def base():
    import motr.controllers.base

    return motr.controllers.base


@pytest.fixture(scope="session")
def api():
    import motr.api

    return motr.api


@pytest.fixture(scope="session")
def io():
    import motr._api.actions.io

    return motr._api.actions.io


@pytest.fixture(scope="session")
def cmd():
    import motr._api.actions.cmd

    return motr._api.actions.cmd


@pytest.fixture(scope="session")
def mkdir():
    import motr._api.actions.mkdir

    return motr._api.actions.mkdir


@pytest.fixture(scope="session")
def write_bytes():
    import motr._api.actions.write_bytes

    return motr._api.actions.write_bytes



import pathlib
import typing

import attr
import trio

import motr._api.actions.io
import motr._api.requirements.action
import motr._api.requirements.requirements
import motr._api.requirements.target
import motr.core.result


@attr.dataclass(frozen=True)
class Mkdir:
    path: pathlib.Path

    async def __call__(
        self,
    ) -> typing.Tuple[motr.core.result.Result, typing.Mapping[str, str]]:
        try:
            await trio.Path(self.path).mkdir(parents=True, exist_ok=True)
        except Exception:
            return motr.core.result.Result.ABORTED, {}
        return motr.core.result.Result.PASSED, {}


def mkdir(
    path: motr._api.actions.io.Input,
) -> motr._api.requirements.requirements.Requirements:
    action = Mkdir(path.path)
    yield from motr._api.requirements.action.action(action)
    yield from motr._api.requirements.target.target(path.path, action)



import motr._api.requirements.requirements
import motr.core.registry
import motr.core.target
import motr.core.target_name


def name_target(
    coerce_to_target: motr.core.target.CoerceToTarget, name: str
) -> motr._api.requirements.requirements.Requirements:
    target = motr.core.target.coerce(coerce_to_target)
    target_name = motr.core.target_name.coerce(name)
    yield motr.core.registry.TargetName(target, target_name)



"""MOTR conditional formatting module."""

import typing

import attr
import blessed.formatters
import cement


def maybe_format(value: str, func: typing.Callable[[str], str]) -> str:
    if value:
        return func(value)
    return value


def maybe_header(value: str, header: str) -> str:
    return maybe_format(value, header.__add__)


@attr.dataclass(frozen=True)
class Color:
    app: cement.App

    def __call__(self, value: str, color: str) -> str:
        assert color in blessed.formatters.COLORS  # type: ignore
        return getattr(self.app.term, color)(value)


@attr.dataclass(frozen=True)
class MaybeColor:
    app: cement.App

    def __call__(self, value: str, color: str) -> str:
        # blessed messed up their type annotations
        assert color in blessed.formatters.COLORS  # type: ignore
        return maybe_format(value, getattr(self.app.term, color))


def register_filter(
    name: str, filter_func: typing.Callable
) -> typing.Callable:
    def register(app: cement.App) -> None:
        assert app.output
        assert app.output.Meta.label == "jinja2"
        app.output.templater.env.filters[name] = filter_func

    return register


COLOR_ARGS = {
    "auto": False,
    "always": True,
    "never": None,
}


def make_term(app: cement.App) -> None:
    assert app.pargs is not None
    color: bool
    # Get it together, blessed.
    color = COLOR_ARGS[app.pargs.color]  # type: ignore
    app.term = blessed.Terminal(force_styling=color)


def add_color_arg(app: cement.App) -> None:
    app.add_arg("--color", default="auto", choices=["auto", "always", "never"])


def load(app: cement.App) -> None:
    app.hook.register(
        "post_setup", register_filter("maybe_header", maybe_header)
    )
    app.hook.register(
        "post_setup",
        register_filter("maybe_color", MaybeColor(app)),
    )
    app.hook.register("post_setup", register_filter("color", Color(app)))
    app.hook.register("post_argument_parsing", make_term)
    app.hook.register("post_setup", add_color_arg)



from motr._api.actions.cmd import cmd_ as cmd
from motr._api.actions.io import Input
from motr._api.actions.mkdir import mkdir
from motr._api.actions.write_bytes import write_bytes
from motr._api.build import build
from motr._api.requirements.action import action
from motr._api.requirements.name_target import name_target
from motr._api.requirements.skipped_name import skipped_name
from motr._api.requirements.target import target
from motr.core.target import deleted

__all__ = [
    "action",
    "target",
    "build",
    "cmd",
    "deleted",
    "Input",
    "mkdir",
    "name_target",
    "skipped_name",
    "write_bytes",
]



import pathlib
import types
from unittest import mock

import pytest
import trio.testing


def make_run_process(returncode):
    async def run_process(command, capture_stdout, capture_stderr, check, env):
        return types.SimpleNamespace(
            returncode=returncode, stdout=b"", stderr=b""
        )

    return run_process


def test_import(api):
    assert api


def test_cmd(api, io):
    string = pathlib.Path("string")
    env = pathlib.Path("env")
    for requirement in api.cmd(
        ["str", io.Input(string / "in"), io.Input(string / "out").as_output()],
        "test_cmd",
        env={
            "IN": io.Input(env / "in"),
            "OUT": io.Input(env / "out").as_output(),
        },
    ):
        assert requirement


def test_mkdir(api, io):
    for requirement in api.mkdir(
        io.Input(pathlib.Path("test-path")),
    ):
        assert requirement


def test_write_bytes(api, io):
    for requirement in api.write_bytes(
        io.Input(pathlib.Path("test-path")), b"test-data"
    ):
        assert requirement


@pytest.mark.parametrize(
    "returncode,expected", [(0, "PASSED"), (1, "FAILED"), (2, "ABORTED")]
)
@trio.testing.trio_test
async def test_inner_cmd(cmd, result, returncode, expected):
    test_cmd = cmd.Cmd((), allowed_codes=frozenset([1]))
    with mock.patch("trio.run_process", new=make_run_process(returncode)):
        my_result, mapping = await test_cmd()
    assert result.Result[expected] is my_result


@trio.testing.trio_test
async def test_inner_mkdir_success(mkdir, result, tmp_path):
    test_path = tmp_path / "test-path"
    my_result, mapping = await mkdir.Mkdir(test_path)()
    assert result.Result.PASSED is my_result


@trio.testing.trio_test
async def test_inner_mkdir_failure(mkdir, result, tmp_path):
    test_path = tmp_path / "test-path"
    test_path.touch()
    my_result, mapping = await mkdir.Mkdir(test_path)()
    assert result.Result.ABORTED is my_result


@trio.testing.trio_test
async def test_inner_write_bytes_success(write_bytes, result, tmp_path):
    test_path = tmp_path / "test-path"
    my_result, mapping = await write_bytes.WriteBytes(test_path, b"")()
    assert result.Result.PASSED is my_result


@trio.testing.trio_test
async def test_inner_write_bytes_failure(write_bytes, result, tmp_path):
    test_path = tmp_path / "test-path"
    test_path.mkdir()
    my_result, mapping = await write_bytes.WriteBytes(test_path, b"")()
    assert result.Result.ABORTED is my_result



import pathlib
import re
import sys
import typing

import attr
import trio
import wheel_filename

from motr import api
from motr._api.actions import cmd as cmd_
from motr._api.requirements import action
from motr._api.requirements import target
from motr.core import result as result_


CONSTRAINTS_FILE = api.Input(pathlib.Path(".motr/constraints.txt"))
REQUIREMENTS = api.Input(pathlib.Path(".motr/requirements"))
REQUIREMENTS_FILE = api.Input(REQUIREMENTS.path.with_suffix(".txt"))


@attr.dataclass(frozen=True)
class SinglePythonPackage:
    root: pathlib.Path
    src: bool


@attr.dataclass(frozen=True)
class PythonPackage:
    prefix: pathlib.Path
    suffix: pathlib.Path
    src: bool

    @property
    def requirements_file(self):
        return api.Input((REQUIREMENTS / self.suffix).with_suffix(".txt"))


def make_parent(input_file):
    path = input_file.path.parent
    yield from api.mkdir(api.Input(path))


@attr.dataclass(frozen=True)
class Venv:
    root: pathlib.Path
    name: str

    @property
    def root_path(self):
        return self.root / self.name

    @property
    def bin_dir(self):
        return self.root_path / "bin"

    @property
    def pip(self):
        return api.Input(self.bin_dir / "pip")

    def create(self):
        action = yield from api.cmd(
            [sys.executable, "-m", "venv", "--clear", self.root_path],
        )
        yield from target.target(self.bin_dir, action)

    def update(self):
        yield from api.cmd(
            [self.pip.as_output(), "install", "-U", "pip"],
            self.bin_dir,
        )

    def install(self, *pip_args):
        return (yield from api.cmd((self.pip, "install") + pip_args))

    def command(self, command, install_action):
        command_path = self.bin_dir / command
        yield from target.target(command_path, install_action)
        return api.Input(command_path)


dot_dir = pathlib.Path(".motr")
requirements_dir = pathlib.Path("requirements")
reports_dir = pathlib.Path("reports")


def setup_venv(name, command, *pip_args):
    venv = Venv(dot_dir, name)

    yield from venv.create()
    yield from venv.update()
    install_action = yield from venv.install(*pip_args)

    return (yield from venv.command(command, install_action))


@attr.dataclass(frozen=True)
class PackageRoots:
    common_root: pathlib.Path
    roots: typing.FrozenSet[typing.Tuple[str, ...]]


@attr.dataclass(frozen=True)
class PyprojectBuild:
    pyproject_build: api.Input
    package_roots: PackageRoots

    async def __call__(self):
        dist_dir = pathlib.Path(".motr/dist")
        await trio.Path(dist_dir).mkdir()
        constraints = []
        for root in self.package_roots.roots:
            outdir = nested
            srcdir = self.package_roots.common_root
            for segment in root:
                outdir /= segment
                srcdir /= segment
            result = await cmd_.Cmd(
                (
                    str(self.pyproject_build.path),
                    "--outdir",
                    str(dist_dir),
                    str(srcdir)
                )
            )(outdir)
            if result is not result_.Result.PASSED:
                return result_.Result.ABORTED
            build_log = await trio.Path(outdir / "out.txt").read_text()
            wheel_re = re.compile(
                r"^Successfully built .* (\S+\.whl)$", flags=re.MULTILINE,
            )
            build_match = re.search(wheel_re, build_log)
            if build_match is None:
                return result_.Result.ABORTED
            build_path = dist_dir / build_match.group(1)
            package_name = wheel_filename.parse_wheel_filename(build_path).project


@attr.dataclass(frozen=True)
class FlitToRequirements:
    flit: api.Input
    outfile: api.Input

    async def __call__(self):
        result, data = await cmd_.Cmd(
            (
                str(self.flit.path), "build", "--format", "wheel",
            ),
        )()
        if result is not result_.Result.PASSED:
            return result_.Result.ABORTED, {}
        build_log = data["err"]
        wheel_re = r"Built wheel:\s*(\S*)"
        build_match = re.search(wheel_re, build_log)
        if build_match is None:
            return result_.Result.ABORTED
        build_path = pathlib.Path(build_match.group(1))
        package_name = wheel_filename.parse_wheel_filename(build_path).project
        await trio.Path(self.outfile.path).write_text(
            f"{package_name} @ {build_path.resolve().as_uri()}\n"
        )
        return result_.Result.PASSED, {}


def flit(outfile):
    flit_cmd = yield from setup_venv("flit", "flit", "flit")
    flit_to_requirements = FlitToRequirements(flit_cmd, outfile)
    yield from action.action(flit_to_requirements, flit_cmd.path)
    yield from target.target(outfile.path, flit_to_requirements)
    return outfile


def venv_wrapper(name, command, requirements_file, *pip_args):
    return (
        yield from setup_venv(
            name,
            command,
            "-r",
            (requirements_dir / requirements_file).with_suffix(".txt"),
            *pip_args,
        )
    )


def pytest_suffix(junit_file):
    return (
        "tests",
        "-p",
        "no:cacheprovider",
        "--junitxml",
        junit_file.as_output(),
    )


def run_pytest(
    env_name, requirement_name, runner, report_name, test_extra_inputs=()
):
    command_name = runner[0]
    requirements_file = yield from flit(REQUIREMENTS_FILE)
    command = yield from venv_wrapper(
        env_name, command_name, requirement_name, "-r", requirements_file
    )
    runner = (command,) + runner[1:]
    report_dir = api.Input(reports_dir / report_name)
    yield from api.mkdir(report_dir)
    junit_file = api.Input(report_dir.path / "junit.xml")
    pytest_action = yield from (
        api.cmd(
            runner + pytest_suffix(junit_file),
            report_dir.path,
            *test_extra_inputs,
            allowed_codes=[1],
        )
    )
    junit2html = yield from venv_wrapper(
        "junit-report", "junit2html", "junit-report"
    )
    html_file = api.Input(report_dir.path / "index.html")
    yield from (
        api.cmd(
            [
                junit2html,
                junit_file,
                html_file.as_output(env_name),
            ],
        )
    )
    return pytest_action


def changes():
    flake8_report_dir = api.Input(reports_dir / "flake8")
    flake8_report_file = flake8_report_dir.path / "index.html"
    flake8 = yield from venv_wrapper("check", "flake8", "check")
    yield from api.mkdir(flake8_report_dir)
    check_action = yield from api.cmd(
        [
            flake8,
            "--format=html",
            "--htmldir",
            flake8_report_dir,
            "--isort-show-traceback",
            "src",
            "tests",
        ],
        allowed_codes=[1],
    )
    yield from target.target(flake8_report_file, check_action, "check")

    requirements_file = yield from flit(REQUIREMENTS_FILE)

    mypy_report_dir = reports_dir / "mypy"
    mypy_report_file = api.Input(mypy_report_dir / "junit.xml")
    mypy_coverage_dir = api.Input(reports_dir / "mypy-coverage")
    mypy_coverage_file = mypy_coverage_dir.path / "index.html"
    mypy = yield from venv_wrapper("mypy", "mypy", "mypy", "-r", requirements_file)
    yield from api.mkdir(api.Input(mypy_report_dir))
    yield from api.mkdir(mypy_coverage_dir)
    mypy_action = yield from api.cmd(
        [
            mypy,
            "--html-report",
            mypy_coverage_dir,
            "--junit-xml",
            mypy_report_file.as_output(),
            "src",
        ],
        mypy_report_dir,
        allowed_codes=[1],
    )
    yield from target.target(mypy_coverage_file, mypy_action, "mypy")
    junit2html = yield from venv_wrapper(
        "junit-report", "junit2html", "junit-report"
    )
    mypy_report_html = api.Input(mypy_report_dir / "index.html")
    yield from api.cmd(
        [junit2html, mypy_report_file, mypy_report_html.as_output("mypy")],
    )

    yield from run_pytest(
        "nocov", "pytest", ("python", "-m", "pytest"), "pytest"
    )

    coverage = yield from venv_wrapper("coverage", "coverage", "coverage")
    coverage_deleted = api.deleted(pathlib.Path(".coverage"))
    erase_action = yield from api.cmd([coverage, "erase"])
    yield from target.target(coverage_deleted, erase_action)
    initial_coverage = "initial-coverage"
    updated_coverage_file = pathlib.Path(".coverage")
    pytest_cover_action = yield from run_pytest(
        "cover",
        "cover",
        ("coverage", "run", "-m", "pytest"),
        "pytest-cover",
        (coverage_deleted,),
    )
    yield from target.target(initial_coverage, pytest_cover_action)
    combine_action = yield from api.cmd(
        [coverage, "combine"],
        initial_coverage,
    )
    yield from target.target(updated_coverage_file, combine_action)
    yield from api.cmd(
        [
            coverage,
            "html",
            "--show-contexts",
            "-d",
            api.Input(reports_dir / "coverage").as_output("cover"),
        ],
        updated_coverage_file,
    )
    coverage_report_action = yield from api.cmd(
        [
            coverage,
            "report",
            "--skip-covered",
            "-m",
            "--fail-under=100",
        ],
        updated_coverage_file,
        allowed_codes=[2],
    )

    yield from target.target("coverage-report", coverage_report_action, "cover")

    profile_dir = api.Input(reports_dir / "profile")

    yield from api.mkdir(profile_dir)
    yield from run_pytest(
        "profile",
        "profile",
        (
            "pyinstrument",
            "--renderer",
            "html",
            "--outfile",
            api.Input(profile_dir.path / "index.html").as_output("profile"),
            "-m",
            "pytest",
        ),
        "pytest-profile",
        (profile_dir.path,),
    )


MOTR_CONFIG = api.build(changes())



from __future__ import annotations

import pathlib
import typing

import attr
import pyrsistent


@attr.dataclass(frozen=True)
class Output:
    path: pathlib.Path
    names: pyrsistent.PVector[str] = pyrsistent.pvector()


@attr.dataclass(frozen=True)
class Input:
    path: pathlib.Path

    def as_output(self, *args: str) -> Output:
        return Output(self.path, pyrsistent.pvector(args))


IOType = typing.Union[Input, Output]

IO = Input, Output



from cement import App
from cement import init_defaults

import motr.controllers.base

# configuration defaults
CONFIG = init_defaults("motr")
CONFIG["motr"]["foo"] = "bar"


class MOTR(App):
    """Max's Obvious Task Runner primary application."""

    class Meta:
        label = "motr"

        # configuration defaults
        config_defaults = CONFIG

        # call sys.exit() on close
        exit_on_close = True

        # load additional framework extensions
        extensions = [
            "yaml",
            "colorlog",
            "jinja2",
            "motr.ext.ext_maybe_fmt",
        ]

        # configuration handler
        config_handler = "yaml"

        # configuration file suffix
        config_file_suffix = ".yml"

        # set the log handler
        log_handler = "colorlog"

        # set the output handler
        output_handler = "jinja2"

        # register handlers
        handlers = [motr.controllers.base.Base]



import pytest


def test_import(main):
    assert main


def test_main_basic(main, test_helpers):
    def run():
        pass

    app = test_helpers.MOTRTest()
    app.run = run
    main.main(lambda: app)
    assert app.exit_code == 0


@pytest.mark.parametrize("debug", (False, True))
def test_assert(main, test_helpers, debug):
    def run():
        assert False

    app = test_helpers.MOTRTest(debug=debug)
    app.run = run
    main.main(lambda: app)
    assert app.exit_code == 1


@pytest.mark.parametrize("debug", (False, True))
def test_error(main, test_helpers, exc, debug):
    def run():
        raise exc.MOTRError("Test exception")

    app = test_helpers.MOTRTest(debug=debug)
    app.run = run
    main.main(lambda: app)
    assert app.exit_code == 1


def test_signal(main, test_helpers):
    def run():
        raise main.CaughtSignal(-1, None)

    app = test_helpers.MOTRTest()
    app.run = run
    main.main(lambda: app)
    assert app.exit_code == 0



import contextlib
import pathlib
import traceback
import typing
from os import chdir  # Allow for easier monkey-patching in testing.

import attr
import tqdm
import trio
from cement import Controller
from cement.utils.version import get_version_banner

import motr
import motr.core.exc
import motr.core.registry
import motr.core.result
import motr.core.runner
import motr.core.target_name

VERSION_BANNER = """
An Opinionated Task Runner %s
%s
""" % (
    motr.__version__,
    get_version_banner(),
)

MOTRFILE_NAME = "motrfile.py"

ResultList = typing.List[
    typing.Tuple[motr.core.runner.RuntimeAction, typing.Mapping[str, str]]
]


@attr.dataclass(frozen=True)
class TQDM:
    pbar: tqdm.tqdm

    @contextlib.contextmanager
    def __call__(
        self,
        action: motr.core.runner.RuntimeAction,
    ) -> typing.Iterator[None]:
        self.pbar.total += 1
        self.pbar.update(0)
        # Do not wrap, because we *want* to skip "cleanup" on failure.
        yield
        self.pbar.update()


def path_from_arg() -> pathlib.Path:
    cwd_path = pathlib.Path(MOTRFILE_NAME).resolve()
    if cwd_path.exists():
        return cwd_path
    for pwd in pathlib.Path(".").resolve().parents:
        motrfile = pwd / MOTRFILE_NAME
        if motrfile.exists():
            return motrfile
    return cwd_path


def motr_open(motrfile: pathlib.Path) -> typing.BinaryIO:
    try:
        return open(motrfile, mode="rb")
    except FileNotFoundError:
        raise motr.core.exc.MOTRError(f"MOTRfile {motrfile} not found.")
    except IsADirectoryError:
        raise motr.core.exc.MOTRError(
            f"MOTRfile at {motrfile} is a directory."
        )
    except PermissionError:
        raise motr.core.exc.MOTRError(
            f"Insufficient permissions to read MOTRfile at {motrfile}."
        )
    except OSError:
        raise motr.core.exc.MOTRError(
            f"Unexpected IO error reading MOTRfile at {motrfile}."
            " Please file an issue."
        )
    except Exception:
        raise motr.core.exc.MOTRError(
            f"Unexpected general error reading MOTRfile at {motrfile}."
            " Please file an issue."
        )


class Base(Controller):
    class Meta:
        label = "base"

        # text displayed at the top of --help output
        description = "An Opinionated Task Runner"

        # text displayed at the bottom of --help output
        epilog = "Usage: motr"

        # controller level arguments. ex: 'motr --version'
        arguments = [
            # add a version banner
            (
                ["-v", "--version"],
                {"action": "version", "version": VERSION_BANNER},
            ),
            (["-l", "--list-targets", "--list"], {"action": "store_true"}),
            (
                ["-t", "--target", "-s", "-e"],
                {"action": "append", "type": motr.core.target_name.TargetName},
            ),
        ]

    def _default(self) -> None:
        """Default action if no sub-command is passed."""

        motrfile = path_from_arg()
        # Hardcode the path for now
        motrdir = motrfile.parent
        chdir(motrdir)
        with motr_open(motrfile) as fh:
            motrfile_text = fh.read()
        try:
            compiled_code = compile(motrfile_text, motrfile, "exec")
        except (SyntaxError, ValueError):
            raise motr.core.exc.MOTRError(
                f"Could not parse MOTRfile at {motrfile};"
                " encountered the following exception:"
                f"\n{traceback.format_exc()}"
            )
        motr_globals: typing.Dict[str, typing.Any] = {}
        try:
            exec(compiled_code, motr_globals)
        except Exception:
            raise motr.core.exc.MOTRError(
                f"The MOTRfile at {motrfile}"
                " encountered the following exception:"
                f"\n{traceback.format_exc()}"
            )
        if "MOTR_CONFIG" not in motr_globals:
            raise motr.core.exc.MOTRError(
                f"The MOTRfile at {motrfile}"
                " did not produce a configuration object."
            )
        motr_config = motr_globals["MOTR_CONFIG"]
        if not isinstance(motr_config, motr.core.registry.Registry):
            raise motr.core.exc.MOTRError(
                f"The MOTRfile at {motrfile} produced an invalid "
                f"configuration object: {motr_config}"
            )
        default, non_default = motr_config.default_and_non_default()
        if self.app.pargs.list_targets:
            self.app.render(
                {"default": default, "non_default": non_default}, "list.jinja2"
            )
            return
        target_names = default
        if self.app.pargs.target is not None:
            target_names = self.app.pargs.target
        targets = motr_config.chosen_targets(target_names)
        passed: ResultList = []
        failed: ResultList = []
        aborted: ResultList = []
        results: typing.Dict[motr.core.result.Result, ResultList] = {
            motr.core.result.Result.PASSED: passed,
            motr.core.result.Result.FAILED: failed,
            motr.core.result.Result.ABORTED: aborted,
        }
        with tqdm.tqdm(total=0) as pbar:
            target_runner = motr.core.runner.Target(
                motr_config, TQDM(pbar), results
            )
            trio.run(motr.core.runner.run_all, target_runner, targets)
        self.app.render(
            {
                "passed": sorted([str(item[0]) for item in passed]),
                "failed": sorted([str(item[0]) for item in failed]),
                "aborted": sorted([str(item[0]) for item in aborted]),
            },
            "results.jinja2",
        )



import motr._api.requirements.requirements
import motr.core.registry
import motr.core.runner
import motr.core.target


def action(
    runtime_action: motr.core.runner.RuntimeAction,
    *args: motr.core.target.CoerceToTarget,
) -> motr._api.requirements.requirements.Requirements:
    yield motr.core.registry.Action(runtime_action)
    for arg in args:
        target = motr.core.target.coerce(arg)
        yield motr.core.registry.ActionInput(runtime_action, target)



import pathlib
import typing

import attr
import trio

import motr._api.actions.io
import motr._api.requirements.action
import motr._api.requirements.requirements
import motr._api.requirements.target
import motr.core.result


@attr.dataclass(frozen=True)
class WriteBytes:
    path: pathlib.Path
    data: bytes  # Put the responsibility for encoding on the caller.

    async def __call__(
        self,
    ) -> typing.Tuple[motr.core.result.Result, typing.Mapping[str, str]]:
        try:
            await trio.Path(self.path).write_bytes(self.data)
        except Exception:
            # Sure would be nice to put traceback information somewhere.
            return motr.core.result.Result.ABORTED, {}
        return motr.core.result.Result.PASSED, {}


def write_bytes(
    path: motr._api.actions.io.Input, data: bytes
) -> motr._api.requirements.requirements.Requirements:
    action = WriteBytes(path.path, data)
    yield from motr._api.requirements.action.action(action)
    yield from motr._api.requirements.target.target(path.path, action)



import enum


class Result(enum.Enum):
    PASSED = False, False
    FAILED = True, False
    ABORTED = True, True

    @property
    def failed(self) -> bool:
        return self.value[0]

    @property
    def aborted(self) -> bool:
        return self.value[1]



from cement import TestApp

import motr.motr_app


class MOTRTest(TestApp, motr.motr_app.MOTR):
    """A sub-class of MOTR that is better suited for testing."""

    class Meta:
        label = "motr"



from __future__ import annotations

import pathlib
import types
import typing

import attr
import pyrsistent
import trio

import motr._api.actions.io
import motr._api.requirements.action
import motr._api.requirements.requirements
import motr._api.requirements.target
import motr.core.result
import motr.core.target

CmdArg = typing.Union[motr._api.actions.io.IOType, str]


@attr.dataclass(frozen=True)
class Cmd:
    cmd: typing.Tuple[str, ...]
    env: pyrsistent.PMap[str, str] = pyrsistent.pmap()
    allowed_codes: pyrsistent.PSet[int] = pyrsistent.pset()

    async def __call__(
        self,
    ) -> typing.Tuple[motr.core.result.Result, typing.Mapping[str, str]]:
        cmd_result = await trio.run_process(
            self.cmd,
            capture_stdout=True,
            capture_stderr=True,
            check=False,
            env=dict(self.env),
        )
        returncode = cmd_result.returncode
        failed = bool(returncode)
        return motr.core.result.Result(
            (failed, failed and returncode not in self.allowed_codes)
        ), {
            "out": cmd_result.stdout.decode(),
            "err": cmd_result.stderr.decode(),
        }


def cmd_(
    cmd: typing.Sequence[CmdArg],
    *inputs: motr.core.target.CoerceToTarget,
    allowed_codes: typing.Collection[int] = frozenset(),
    env: typing.Mapping[str, CmdArg] = types.MappingProxyType({})
) -> motr._api.requirements.requirements.Requirements:
    action = Cmd(
        tuple(
            str(arg.path) if isinstance(arg, motr._api.actions.io.IO) else arg
            for arg in cmd
        ),
        allowed_codes=pyrsistent.pset(allowed_codes),
        env=pyrsistent.pmap(
            {
                key: str(val.path)
                if isinstance(val, motr._api.actions.io.IO)
                else val
                for key, val in env.items()
            }
        ),
    )
    yield from motr._api.requirements.action.action(
        action, *inputs, *extra_inputs(cmd), *extra_inputs(env.values())
    )
    yield from extra_targets(action, cmd)
    yield from extra_targets(action, env.values())
    return action


def extra_inputs(
    cmd_args: typing.Iterable[CmdArg],
) -> typing.Iterator[pathlib.Path]:
    for cmd_arg in cmd_args:
        if isinstance(cmd_arg, motr._api.actions.io.Input):
            yield cmd_arg.path


def extra_targets(
    action: motr.core.runner.RuntimeAction,
    cmd_args: typing.Iterable[CmdArg],
) -> motr._api.requirements.requirements.Requirements:
    for cmd_arg in cmd_args:
        if isinstance(cmd_arg, motr._api.actions.io.Output):
            yield from motr._api.requirements.target.target(
                cmd_arg.path, action, *cmd_arg.names
            )



from typing import Callable

from cement.core.exc import CaughtSignal

import motr.core.exc
import motr.motr_app


def main(
    app_class: Callable[[], motr.motr_app.MOTR] = motr.motr_app.MOTR
) -> None:
    with app_class() as app:
        try:
            app.run()

        except AssertionError as e:
            print("AssertionError > %s" % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback

                traceback.print_exc()

        except motr.core.exc.MOTRError as e:
            print("MOTRError > %s" % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback

                traceback.print_exc()

        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print("\n%s" % e)
            app.exit_code = 0



import contextlib
import itertools

import attr
import pyrsistent
import pytest
import trio

EMPTY_MAP = pyrsistent.pmap()


@attr.dataclass(frozen=True)
class null_action:
    result: object

    async def __call__(self):
        return self.result


@attr.dataclass(frozen=True)
class SynchAction:
    result: object  # Filled in by test.
    before: trio.Event
    after: trio.Event

    async def __call__(self):
        await self.before.wait()
        self.after.set()
        return self.result


@contextlib.contextmanager
def null_report(action):
    yield


def test_import(runner):
    assert runner


def test_task_wrapper(runner):
    wrapper = runner.TaskWrapper()
    run_count = 0

    async def run():
        nonlocal run_count
        run_count += 1

    trio.run(wrapper, run)
    assert run_count == 1
    trio.run(wrapper, run)
    assert run_count == 1


@pytest.mark.parametrize("count", list(range(10)))
def test_run_all(runner, count):
    run_count = 0

    async def run(arg):
        nonlocal run_count
        run_count += 1

    trio.run(runner.run_all, run, [None] * count)
    assert run_count == count


def test_target_action_comprehensive(exc, result, runner, api):
    pass_result = result.Result.PASSED
    fail_result = result.Result.FAILED
    abort_result = result.Result.ABORTED
    first_action = null_action((pass_result, EMPTY_MAP))
    second_action = null_action((abort_result, EMPTY_MAP))
    third_action = null_action((pass_result, EMPTY_MAP.set(1, 1)))
    reg = api.build(
        itertools.chain(
            api.action(first_action),
            api.target("first_res", first_action),
            api.action(second_action, "first_res"),
            api.target("second_res", second_action),
            api.action(third_action, "second_res"),
            api.target("third_res", third_action),
        ),
    )
    results = {pass_result: [], fail_result: [], abort_result: []}
    target = runner.Target(reg, null_report, results)

    with pytest.raises(exc.MOTRTaskError):
        trio.run(target, "third_res")

    assert results == {
        pass_result: [(first_action, EMPTY_MAP)],
        fail_result: [],
        abort_result: [(second_action, EMPTY_MAP)],
    }


@pytest.mark.parametrize(
    "sequence",
    [
        (0, 0, 1, 1),
        (0, 1, 0, 1),
        (0, 1, 1, 0),
        (1, 0, 0, 1),
        (1, 0, 1, 0),
        (1, 1, 0, 0),
    ],
)
def test_implicit_sequencing(result, runner, api, sequence):
    event = trio.Event()
    event.set()
    sequences = [[], []]
    pass_result = result.Result.PASSED
    fail_result = result.Result.FAILED
    abort_result = result.Result.ABORTED
    final_action = null_action((pass_result, EMPTY_MAP))

    for index in sequence:
        new_event = trio.Event()
        sequences[index].append(
            SynchAction((pass_result, EMPTY_MAP), event, new_event)
        )
        event = new_event

    args = []

    for index, sub_sequence in enumerate(sequences):
        action_0 = f"{index}0"
        res_0 = action_0 + "_res"
        action_1 = f"{index}1"
        res_1 = action_1 + "_res"
        args.extend(api.action(sub_sequence[0]))
        args.extend(api.target(res_0, sub_sequence[0]))
        args.extend(api.action(sub_sequence[1], res_0))
        args.extend(api.target(res_1, sub_sequence[1]))

    args.extend(api.action(final_action, "01_res", "11_res"))
    args.extend(api.target("final_res", final_action))
    reg = api.build(args)
    results = {pass_result: [], fail_result: [], abort_result: []}
    target = runner.Target(reg, null_report, results)
    trio.run(target, "final_res")

    assert results[pass_result]
    assert not results[fail_result]
    assert not results[abort_result]



import motr._api.requirements.name_target
import motr._api.requirements.requirements
import motr.core.registry
import motr.core.runner
import motr.core.target


def target(
    coerce_to_target: motr.core.target.CoerceToTarget,
    parent: motr.core.runner.RuntimeAction,
    *args: str
) -> motr._api.requirements.requirements.Requirements:
    target = motr.core.target.coerce(coerce_to_target)
    yield motr.core.registry.ActionOutput(parent, target)
    for arg in args:
        yield from motr._api.requirements.name_target.name_target(
            coerce_to_target, arg
        )



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['motr',
 'motr._api',
 'motr._api.actions',
 'motr._api.requirements',
 'motr.controllers',
 'motr.core',
 'motr.ext',
 'motr.plugins',
 'motr.templates']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

install_requires = \
['attrs',
 'blessed',
 'cement~=3.0, >=3.0.4',
 'colorlog',
 'jinja2',
 'pyrsistent',
 'python-slugify',
 'pyyaml',
 'tqdm',
 'trio']

entry_points = \
{'console_scripts': ['motr = motr.main:main']}

setup(name='motr',
      version='0.1.3',
      description="Max's Obvious Task Runner",
      author='Max Woerner Chase',
      author_email='max.chase@gmail.com',
      url=None,
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.8',
     )



import pathlib

import nox


def install_from_requirements(session, filename, project=None):
    args = ["-r", f"requirements/{filename}.txt"]
    if project is not None:
        args.append(project)
    session.install(*args)


REPORTS = pathlib.Path("reports")
PYTEST = "pytest"
COVER = "pytest-cover"
PROFILE = "pytest-profile"
PYSPY = "pytest-py-spy"


def junit_xml_root(name):
    return REPORTS / name


@nox.session
def clean(session):
    install_from_requirements(session, "coverage")
    session.run("coverage", "erase")


@nox.session
def check(session):
    install_from_requirements(session, "check")
    report_dir = REPORTS / "flake8"
    report_dir.mkdir(parents=True, exist_ok=True)
    session.run(
        "flake8",
        "--format=html",
        f"--htmldir={report_dir}",
        "--isort-show-traceback",
        "src",
        "tests",
    )


@nox.session
def mypy(session):
    install_from_requirements(session, "mypy", ".")
    report_dir = REPORTS / "mypy"
    report_dir.mkdir(parents=True, exist_ok=True)
    report_file = report_dir / "junit.xml"
    coverage_report_dir = REPORTS / "mypy-coverage"
    coverage_report_dir.mkdir(parents=True, exist_ok=True)
    session.run(
        "mypy",
        f"--html-report={coverage_report_dir}",
        f"--junit-xml={report_file}",
        "src",
    )


@nox.session
def mypy_report(session):
    install_from_requirements(session, "junit-report")
    report_dir = REPORTS / "mypy"
    xml = str(report_dir / "junit.xml")
    html = str(report_dir / "index.html")
    session.run("junit2html", xml, html)


@nox.session
def nocov(session):
    install_from_requirements(session, "pytest", ".")
    report_dir = junit_xml_root(PYTEST)
    report_dir.mkdir(parents=True, exist_ok=True)
    junit_xml = report_dir / "junit.xml"
    session.run(
        "python",
        "-m",
        "pytest",
        "tests",
        "-p",
        "no:cacheprovider",
        f"--junitxml={junit_xml}",
    )


@nox.session
def cover(session):
    install_from_requirements(session, "cover", ".")
    report_dir = junit_xml_root(COVER)
    report_dir.mkdir(parents=True, exist_ok=True)
    junit_xml = report_dir / "junit.xml"
    session.run(
        "coverage",
        "run",
        "-m",
        "pytest",
        "tests",
        "-p",
        "no:cacheprovider",
        f"--junitxml={junit_xml}",
    )


@nox.session
def report(session):
    install_from_requirements(session, "report")
    session.run("coverage", "combine")
    # I am skeptical of the organization; do not rely on it.
    # session.run("limit-coverage")
    session.run(
        "coverage", "html", "--show-contexts", "-d", "reports/coverage"
    )
    session.run(
        "coverage", "report", "--skip-covered", "-m", "--fail-under=100"
    )


@nox.session
def profile(session):
    install_from_requirements(session, "profile", ".")
    report_dir = junit_xml_root(PROFILE)
    report_dir.mkdir(parents=True, exist_ok=True)
    junit_xml = report_dir / "junit.xml"
    html_dir = REPORTS / "profile"
    html_dir.mkdir(parents=True, exist_ok=True)
    outfile = html_dir / "index.html"
    session.run(
        "pyinstrument",
        "--renderer",
        "html",
        f"--outfile={outfile}",
        "-m",
        "pytest",
        "tests",
        "-p",
        "no:cacheprovider",
        f"--junitxml={junit_xml}",
    )


@nox.session
def pyspy(session):
    install_from_requirements(session, "pyspy", ".")
    report_dir = junit_xml_root(PYSPY)
    report_dir.mkdir(parents=True, exist_ok=True)
    junit_xml = report_dir / "junit.xml"
    svg_dir = REPORTS / "py-spy"
    svg_dir.mkdir(parents=True, exist_ok=True)
    outfile = svg_dir / "profile.svg"
    session.run(
        "py-spy",
        "record",
        "-o",
        str(outfile),
        "--",
        "python",
        "-m",
        "pytest",
        "tests",
        "-p",
        "no:cacheprovider",
        f"--junitxml={junit_xml}",
    )


@nox.session
@nox.parametrize("pytest_kind", [PYTEST, COVER, PROFILE, PYSPY])
def pyest_report(session, pytest_kind):
    install_from_requirements(session, "junit-report")
    junit_xml_dir = junit_xml_root(pytest_kind)
    junit_xml = str(junit_xml_dir / "junit.xml")
    html = str(REPORTS / pytest_kind / "index.html")
    session.run("junit2html", junit_xml, html)



from __future__ import annotations

import functools
import typing

import attr
import pyrsistent

if typing.TYPE_CHECKING:
    import motr.core.runner
    import motr.core.target
    import motr.core.target_name


@attr.dataclass(frozen=True)
class Action:
    action: motr.core.runner.RuntimeAction


@attr.dataclass(frozen=True)
class ActionInput:
    action: motr.core.runner.RuntimeAction
    target: motr.core.target.Target


@attr.dataclass(frozen=True)
class ActionOutput:
    parent: motr.core.runner.RuntimeAction
    target: motr.core.target.Target


@attr.dataclass(frozen=True)
class TargetName:
    target: motr.core.target.Target
    name: motr.core.target_name.TargetName


@attr.dataclass(frozen=True)
class SkippedName:
    name: motr.core.target_name.TargetName


Requirement = typing.Union[
    Action, ActionInput, ActionOutput, TargetName, SkippedName
]


@attr.dataclass(frozen=True)
class ActionData:
    inputs: pyrsistent.PSet[motr.core.target.Target] = pyrsistent.pset()

    def add_input(self, target: motr.core.target.Target) -> ActionData:
        return attr.evolve(self, inputs=self.inputs.add(target))


@attr.dataclass(frozen=True)
class Registry:
    # These are the user-selectable targets.
    # A single name can refer to multiple targets.
    _choosable_targets: pyrsistent.PMap[
        motr.core.target_name.TargetName,
        pyrsistent.PSet[motr.core.target.Target],
    ] = pyrsistent.pmap()
    # These are the defined actions.
    _actions: pyrsistent.PMap[
        motr.core.runner.RuntimeAction, ActionData
    ] = pyrsistent.pmap()
    # This is the mapping from targets to prerequisites.
    # A single action can satisfy many targets.
    _parent_action: pyrsistent.PMap[
        motr.core.target.Target, motr.core.runner.RuntimeAction
    ] = pyrsistent.pmap()
    # Some user-selectable targets are not selected by default.
    # That is indicated with this set.
    _default_skips: pyrsistent.PSet[
        motr.core.target_name.TargetName
    ] = pyrsistent.pset()
    _last_added: typing.Optional[motr.core.runner.RuntimeAction] = None

    def inputs(
        self, action: motr.core.runner.RuntimeAction
    ) -> pyrsistent.PSet[motr.core.target.Target]:
        return self._actions[action].inputs

    def chosen_targets(
        self, names: typing.Iterable[motr.core.target_name.TargetName]
    ) -> typing.Set[motr.core.target.Target]:
        return {
            target
            for name in names
            for target in self._choosable_targets[name]
        }

    def parent(
        self, target: motr.core.target.Target
    ) -> motr.core.runner.RuntimeAction:
        return self._parent_action[target]

    def default_and_non_default(
        self,
    ) -> typing.Tuple[
        typing.List[motr.core.target_name.TargetName],
        typing.List[motr.core.target_name.TargetName],
    ]:
        return (
            sorted(
                set(self._choosable_targets).difference(self._default_skips)
            ),
            sorted(self._default_skips),
        )

    def _require_action(self, requirement: Action) -> Registry:
        if requirement.action in self._actions:
            return self
        return attr.evolve(
            self,
            actions=self._actions.set(requirement.action, ActionData()),
            last_added=requirement.action,
        )

    def _require_action_input(self, requirement: ActionInput) -> Registry:
        if requirement.action not in self._actions:
            raise ValueError(
                f"Cannot add target {requirement.target!r} as an input to"
                f" action {requirement.action!r}. Action"
                f" {requirement.action!r} has not been added to the registry."
            )
        if requirement.target not in self._parent_action:
            raise ValueError(
                f"Cannot add target {requirement.target!r} as an input to"
                f" action {requirement.action!r}. Target"
                f" {requirement.target!r} has not been added to the registry."
            )
        if self._parent_action[requirement.target] == requirement.action:
            raise ValueError(
                f"Cannot add target {requirement.target!r} as an input to"
                f" action {requirement.action!r}. The parent action of target"
                f" {requirement.target!r} is {requirement.action!r}."
            )
        if requirement.target in self._actions[requirement.action].inputs:
            return self
        if requirement.action != self._last_added:
            raise ValueError(
                f"Cannot add target {requirement.target!r} as an input to"
                f" action {requirement.action!r}. The most recently added"
                f" action is {self._last_added!r}."
            )
        return attr.evolve(
            self,
            actions=self._actions.set(
                requirement.action,
                self._actions[requirement.action].add_input(
                    requirement.target
                ),
            ),
        )

    def _require_action_output(self, requirement: ActionOutput) -> Registry:
        if (
            self._parent_action.get(requirement.target, requirement.parent)
            != requirement.parent
        ):
            raise ValueError(
                f"Cannot add target {requirement.target!r} with parent action"
                f" {requirement.parent!r}."
                f" Target {requirement.target!r} already has parent action"
                f" {self._parent_action[requirement.target]!r}."
            )
        if requirement.parent not in self._actions:
            raise ValueError(
                f"Cannot add target {requirement.target!r} with parent action"
                f" {requirement.parent!r}."
                f" Parent action {requirement.parent!r} has not been added to"
                " the registry."
            )
        return attr.evolve(
            self,
            parent_action=self._parent_action.set(
                requirement.target, requirement.parent
            ),
        )

    def _require_target_name(self, requirement: TargetName) -> Registry:
        if requirement.target not in self._parent_action:
            raise ValueError(
                f"Cannot name target {requirement.target!r}"
                f" {requirement.name!r}."
                f" Target {requirement.target!r} has not been added to the"
                " registry."
            )
        return attr.evolve(
            self,
            choosable_targets=self._choosable_targets.set(
                requirement.name,
                self._choosable_targets.get(
                    requirement.name, pyrsistent.pset()
                ).add(requirement.target),
            ),
        )

    def _require_skipped_name(self, requirement: SkippedName) -> Registry:
        if requirement.name not in self._choosable_targets:
            raise ValueError(
                f"Cannot skip name {requirement.name!r}."
                f" Name {requirement.name!r} has not been assigned to any"
                " target."
            )
        return attr.evolve(
            self, default_skips=self._default_skips.add(requirement.name)
        )

    def require(self, requirement: Requirement) -> Registry:
        raise NotImplementedError

    if not typing.TYPE_CHECKING:  # pragma: no branch
        require = functools.singledispatchmethod(require)
        # We need to specify the exact classes, because relying on annotations
        # doesn't work.
        # This is because it resolves the *entire* annotation, even though it
        # arguably doesn't need it, and it can't resolve the return value,
        # because the class isn't bound to the module namespace yet.
        require.register(Action, _require_action)
        require.register(ActionInput, _require_action_input)
        require.register(ActionOutput, _require_action_output)
        require.register(TargetName, _require_target_name)
        require.register(SkippedName, _require_skipped_name)



import re
from unittest import mock

import blessed
import oschmod
import pytest


def fake_open_for_read(exc_type):
    def open(path, mode):
        assert mode == "rb"
        raise exc_type

    return open


def not_found(path):
    return f"^MOTRfile {re.escape(str(path))} not found.$"


def not_a_file(path):
    return f"^MOTRfile at {re.escape(str(path))} is a directory.$"


def no_permission(path):
    return (
        "^Insufficient permissions to read MOTRfile at"
        f" {re.escape(str(path))}.$"
    )


def unknown_os_error(path):
    return (
        f"^Unexpected IO error reading MOTRfile at {re.escape(str(path))}."
        " Please file an issue.$"
    )


def unknown_general_error(path):
    return (
        f"Unexpected general error reading MOTRfile at {re.escape(str(path))}."
        " Please file an issue.$"
    )


def no_parse(path):
    return (
        f"^Could not parse MOTRfile at {re.escape(str(path))};"
        " encountered the following exception:"
    )


def runtime_exception(path):
    return (
        f"^The MOTRfile at {re.escape(str(path))}"
        " encountered the following exception:"
    )


def no_config(path):
    return (
        f"^The MOTRfile at {re.escape(str(path))}"
        " did not produce a configuration object.$"
    )


def invalid_config(path):
    return (
        f"^The MOTRfile at {re.escape(str(path))} "
        "produced an invalid configuration object:"
    )


@pytest.fixture(name="implicit_runner")
def _implicit_runner(test_helpers, exc):
    def implicit_runner(msg):
        with pytest.raises(exc.MOTRError, match=msg):
            with test_helpers.MOTRTest() as app:
                app.run()

    return implicit_runner


def test_no_config_implicit(implicit_expected, implicit_runner):
    msg = not_found(implicit_expected)
    implicit_runner(msg)


def test_not_a_file_implicit(implicit_actual, implicit_runner):
    implicit_actual.mkdir()
    msg = not_a_file(implicit_actual)
    implicit_runner(msg)


def test_no_permission_implicit(implicit_actual, implicit_runner):
    implicit_actual.touch()
    # If this doesn't work on Windows,
    # the first argument needs to be coerced to str.
    oschmod.set_mode(implicit_actual, "a-r")
    msg = no_permission(implicit_actual)
    implicit_runner(msg)


@pytest.mark.parametrize("char", [")", "\0"])
def test_unparseable_config_implicit(char, implicit_actual, implicit_runner):
    implicit_actual.write_text(char)
    msg = no_parse(implicit_actual)
    implicit_runner(msg)


def test_runtime_exception_config_implicit(implicit_actual, implicit_runner):
    implicit_actual.write_text("1/0")
    msg = runtime_exception(implicit_actual)
    implicit_runner(msg)


def test_no_config_obj_implicit(implicit_actual, implicit_runner):
    implicit_actual.touch()
    msg = no_config(implicit_actual)
    implicit_runner(msg)


def test_invalid_config_implicit(implicit_actual, implicit_runner):
    implicit_actual.write_text("MOTR_CONFIG = None")
    msg = invalid_config(implicit_actual)
    implicit_runner(msg)


UNKNOWN_ERROR_PARAMS = pytest.mark.parametrize(
    "msg_writer,patched_open",
    [
        (unknown_os_error, fake_open_for_read(OSError)),
        (unknown_general_error, fake_open_for_read(Exception)),
    ],
)


@UNKNOWN_ERROR_PARAMS
def test_unknown_error_implicit(
    implicit_actual, implicit_runner, msg_writer, patched_open
):
    implicit_actual.touch()
    msg = msg_writer(implicit_actual)
    with mock.patch("motr.controllers.base.open", new=patched_open):
        implicit_runner(msg)


LIST_PARAMS = pytest.mark.parametrize(
    "config,result",
    [
        (
            """\
import itertools

from motr import api


ACTION = object()  # Not much is needed if we're not running the tasks.

MOTR_CONFIG = api.build(
    itertools.chain(
        api.action(ACTION),
        api.target("default_target", ACTION, "default"),
        api.target("non_default_target", ACTION, "non_default"),
        api.skipped_name("non_default"),
    ),
)
""",
            """\
{term.green}Default Targets:
default
{term.normal}{term.yellow}Non-default Targets:
non_default
{term.normal}
""",
        ),
        (
            """\
import itertools

from motr import api


ACTION = object()  # Ditto

MOTR_CONFIG = api.build(
    itertools.chain(
        api.action(ACTION),
        api.target("target", ACTION, "default-actions"),
    ),
)
""",
            """\
{term.green}Default Targets:
default-actions
{term.normal}
""",
        ),
    ],
)

LIST_ARG = pytest.mark.parametrize(
    "list_arg", ["-l", "--list-targets", "--list"]
)
COLOR_ARG = pytest.mark.parametrize(
    "term,color_arg",
    [
        (blessed.Terminal(force_styling=None), "never"),
        (blessed.Terminal(force_styling=True), "always"),
    ],
)


@LIST_ARG
@LIST_PARAMS
@COLOR_ARG
def test_list_implicit(
    implicit_actual, test_helpers, list_arg, term, color_arg, config, result
):
    implicit_actual.write_text(config)

    with test_helpers.MOTRTest(argv=[list_arg, "--color", color_arg]) as app:
        app.run()

    assert app.last_rendered[1] == result.format(term=term)


EMPTY_REGISTRY = """\
from motr import api


MOTR_CONFIG = api.build(())
"""

# Should make it warn on empty registry?
EMPTY_MSG = """\
{term.yellow}No actions ran.{term.normal}
"""


@COLOR_ARG
def test_empty_registry_implicit(
    implicit_actual, test_helpers, term, color_arg
):
    implicit_actual.write_text(EMPTY_REGISTRY)

    with test_helpers.MOTRTest(argv=["--color", color_arg]) as app:
        app.run()

    assert app.last_rendered[1] == EMPTY_MSG.format(term=term)


@COLOR_ARG
def test_empty_registry_implicit_twice(
    implicit_actual, test_helpers, term, color_arg
):
    implicit_actual.write_text(EMPTY_REGISTRY)

    with test_helpers.MOTRTest(argv=["--color", color_arg]) as app:
        app.run()

    assert app.last_rendered[1] == EMPTY_MSG.format(term=term)

    with test_helpers.MOTRTest(argv=["--color", color_arg]) as app:
        app.run()

    assert app.last_rendered[1] == EMPTY_MSG.format(term=term)


BASIC_REGISTRY = """\
import pathlib

from motr import api
from motr._api.actions import io
from motr._api.requirements import name_target

target_file = pathlib.Path("target-file")

def changes():
    yield from api.write_bytes(io.Input(target_file), b"")
    yield from name_target.name_target(target_file, "test")


MOTR_CONFIG = api.build(changes())
"""


BASIC_MSG = """\
{term.green}Passed:
WriteBytes(path=PosixPath('target-file'), data=b''){term.normal}
"""


@COLOR_ARG
def test_basic_registry_implicit(
    implicit_actual, test_helpers, term, color_arg
):
    implicit_actual.write_text(BASIC_REGISTRY)

    with test_helpers.MOTRTest(argv=["--color", color_arg]) as app:
        app.run()

    assert app.last_rendered[1] == BASIC_MSG.format(term=term)


@COLOR_ARG
def test_basic_registry_targets(
    implicit_actual, test_helpers, term, color_arg
):
    implicit_actual.write_text(BASIC_REGISTRY)

    with test_helpers.MOTRTest(
        argv=["--color", color_arg, "--target", "test"]
    ) as app:
        app.run()

    assert app.last_rendered[1] == BASIC_MSG.format(term=term)
